from pathlib import Path
from typing import Any, BinaryIO, TextIO

__version__: str

def load(fp: BinaryIO, /, *, parse_datetime: bool = True) -> dict[str, Any]: ...
def loads(s: str, /, *, parse_datetime: bool = True) -> dict[str, Any]: ...
def dump(obj: Any, file: str | Path | TextIO) -> int: ...
def dumps(obj: Any) -> str: ...

class YAMLDecodeError(ValueError): ...
class YAMLEncodeError(TypeError): ...
